"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""
import uuid
import datetime
import time

from inmanta import data
from inmanta import const
import pytest
import pymongo
import logging
from tornado import gen


class Doc(data.BaseDocument):
    name = data.Field(field_type=str, required=True)
    field1 = data.Field(field_type=str, default=None)
    field2 = data.Field(field_type=bool, default=False)
    field3 = data.Field(field_type=list, default=[])
    field4 = data.Field(field_type=dict, default={})

    __indexes__ = [
        dict(keys=[("name", pymongo.ASCENDING)], unique=True)
    ]


class EnumDoc(data.BaseDocument):
    action = data.Field(field_type=const.ResourceAction, required=True)


@pytest.mark.gen_test
def test_motor(motor):
    yield motor.testCollection.insert_one({"a": 1, "b": "abcd"})
    results = yield motor.testCollection.find_one({"a": {"$gt": 0}})

    assert "_id" in results

    yield motor.testCollection.insert_one({"a": {"b": {"c": 1}}})
    results = motor.testCollection.find({})

    while (yield results.fetch_next):
        assert "a" in results.next_object()


def test_collection_naming():
    assert Doc.collection_name() == "Doc"


def test_ctor():
    # Mongo data should not contain an id
    with pytest.raises(AttributeError):
        Doc(from_mongo=True, id="1234")

    # ID is autogenerated for new objects
    with pytest.raises(AttributeError):
        Doc(from_mongo=True, id="1234")

    # Required attributes have to be set in the ctor
    with pytest.raises(AttributeError):
        Doc()

    # Only defined fields are allowed
    with pytest.raises(AttributeError):
        Doc(test="")

    # A required field cannot be none
    with pytest.raises(TypeError):
        Doc(name=None)

    # Fields require the correct type
    with pytest.raises(TypeError):
        Doc(name=1234)


def test_document_def():
    t = Doc(name="doc")
    with pytest.raises(TypeError):
        t.id = "1234"

    t.id = uuid.uuid4()
    json = t.to_dict()

    assert "id" in json
    assert "_id" in t.to_mongo()


@pytest.mark.gen_test
def test_document_insert(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")

    yield d.insert()
    docs = yield motor.Doc.find({}).to_list(length=10)
    assert len(docs) == 1

    doc = yield Doc.get_by_id(d.id)
    assert doc.name == d.name


@pytest.mark.gen_test
def test_get_by_id(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    yield d.insert()

    d1 = yield Doc.get_by_id(d.id)
    assert d1.name == d.name

    d2 = yield Doc.get_by_id(uuid.uuid4())
    assert d2 is None


@pytest.mark.gen_test
def test_defaults(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    yield d.insert()

    assert d.to_dict()["field1"] is None
    assert not d.field2

    d2 = yield Doc.get_by_id(d.id)
    d2.insert()

    assert not d2.field2

    d.field3.append(1)


@pytest.mark.gen_test
def test_document_update(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    yield d.insert()

    yield d.update(name="test2")
    result = yield motor.Doc.find_one({"name": "test2"})
    assert "name" in result


@pytest.mark.gen_test
def test_document_delete(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    yield d.insert()
    yield Doc.delete_all(name="test")

    docs = yield Doc.get_list()
    assert len(docs) == 0


@pytest.mark.gen_test
def test_key_escape(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    d.field4 = {"a.b": "1", "a$b": "2", "a\\b": "3", "a.b.c": 4}
    yield d.insert()

    d2 = yield Doc.get_by_id(d.id)

    for k in d.field4.keys():
        assert k in d2.field4


@pytest.mark.gen_test
def test_nested_key_escape(motor):
    Doc.set_connection(motor)

    d = Doc(name="test")
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"]["current"] = {}
    d.field4["a0a7f22b-bc71-51f6-bc91-f2b0e368b786"]["changes"]["routes"]["current"]["172.19.0.0/16"] = "1.1.1.1"
    yield d.insert()

    yield Doc.get_by_id(d.id)


@pytest.mark.gen_test
def test_enum_field(motor):
    EnumDoc.set_connection(motor)

    d = EnumDoc(action=const.ResourceAction.deploy)
    yield d.insert()

    new = yield EnumDoc.get_by_id(d.id)

    assert new.action is const.ResourceAction.deploy
    assert new.to_dict()["action"] == const.ResourceAction.deploy


@pytest.mark.gen_test
def test_project(data_module):
    project = data.Project(name="test")
    yield project.insert()

    projects = yield data.Project.get_list(name="test")
    assert len(projects) == 1
    assert projects[0].id == project.id

    other = yield data.Project.get_by_id(project.id)
    assert project != other
    assert project.id == other.id


@pytest.mark.gen_test
def test_project_unique(data_module):
    project = data.Project(name="test")
    yield project.insert()

    project = data.Project(name="test")
    with pytest.raises(pymongo.errors.DuplicateKeyError):
        yield project.insert()


def test_project_no_project_name(data_module):
    with pytest.raises(AttributeError):
        data.Project()


@pytest.mark.gen_test
def test_environment(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()
    assert env.project == project.id

    yield project.delete_cascade()

    projects = yield data.Project.get_list()
    envs = yield data.Environment.get_list()
    assert len(projects) == 0
    assert len(envs) == 0


@pytest.mark.gen_test
def test_environment_no_environment_name(data_module):
    project = data.Project(name="test")
    yield project.insert()
    with pytest.raises(AttributeError):
        data.Environment(project=project.id, repo_url="", repo_branch="")


@pytest.mark.gen_test
def test_environment_no_project_id(data_module):
    project = data.Project(name="test")
    yield project.insert()
    with pytest.raises(AttributeError):
        data.Environment(name="dev", repo_url="", repo_branch="")


@pytest.mark.gen_test
def test_environment_cascade_content_only(data_module):
    project = data.Project(name="proj")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=uuid.uuid4())
    yield agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.id, name="agi1", tid=env.id)
    yield agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.id, name="agi2", tid=env.id)
    yield agi2.insert()

    agent = data.Agent(environment=env.id, name="agi1", last_failover=datetime.datetime.now(), paused=False,
                       primary=agi1.id)
    yield agent.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(version=version, environment=env.id)
    yield cm.insert()

    resource_ids = []
    for i in range(5):
        path = "/etc/file" + str(i)
        key = "std::File[agent1,path=" + path + "]"
        res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
        yield res1.insert()
        resource_ids.append(res1.id)

    code = data.Code(version=version, resource="std::File", environment=env.id)
    yield code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    yield unknown_parameter.insert()

    yield env.set(data.AUTO_DEPLOY, True)

    assert (yield data.Project.get_by_id(project.id)) is not None
    assert (yield data.Environment.get_by_id(env.id)) is not None
    assert (yield data.AgentProcess.get_by_id(agent_proc.id)) is not None
    assert (yield data.AgentInstance.get_by_id(agi1.id)) is not None
    assert (yield data.AgentInstance.get_by_id(agi2.id)) is not None
    assert (yield data.Agent.get_by_id(agent.id)) is not None
    for current_id in resource_ids:
        assert (yield data.Resource.get_by_id(current_id)) is not None
    assert (yield data.Code.get_by_id(code.id)) is not None
    assert (yield data.UnknownParameter.get_by_id(unknown_parameter.id)) is not None
    assert (yield env.get(data.AUTO_DEPLOY)) is True

    yield env.delete_cascade(only_content=True)

    assert (yield data.Project.get_by_id(project.id)) is not None
    assert (yield data.Environment.get_by_id(env.id)) is not None
    assert (yield data.AgentProcess.get_by_id(agent_proc.id)) is None
    assert (yield data.AgentInstance.get_by_id(agi1.id)) is None
    assert (yield data.AgentInstance.get_by_id(agi2.id)) is None
    assert (yield data.Agent.get_by_id(agent.id)) is None
    for current_id in resource_ids:
        assert (yield data.Resource.get_by_id(current_id)) is None
    assert (yield data.Code.get_by_id(code.id)) is None
    assert (yield data.UnknownParameter.get_by_id(unknown_parameter.id)) is None
    assert (yield env.get(data.AUTO_DEPLOY)) is True


@pytest.mark.gen_test
def test_environment_set_setting_parameter(data_module):
    project = data.Project(name="proj")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    assert (yield env.get(data.AUTO_DEPLOY)) is False
    yield env.set(data.AUTO_DEPLOY, True)
    assert (yield env.get(data.AUTO_DEPLOY)) is True
    yield env.unset(data.AUTO_DEPLOY)
    assert (yield env.get(data.AUTO_DEPLOY)) is False

    with pytest.raises(KeyError):
        yield env.set("set_non_existing_parameter", 1)
    with pytest.raises(KeyError):
        yield env.get("get_non_existing_parameter")
    with pytest.raises(AttributeError):
        yield env.set(data.AUTO_DEPLOY, 5)


@pytest.mark.gen_test
def test_agent_process(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    yield agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.id, name="agi1", tid=env.id)
    yield agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.id, name="agi2", tid=env.id)
    yield agi2.insert()

    agent_procs = yield data.AgentProcess.get_by_env(env=env.id)
    assert len(agent_procs) == 1
    assert agent_procs[0].id == agent_proc.id

    assert (yield data.AgentProcess.get_by_sid(sid)).id == agent_proc.id
    assert (yield data.AgentProcess.get_by_sid(uuid.UUID(int=1))) is None

    live_procs = yield data.AgentProcess.get_live()
    assert len(live_procs) == 1
    assert live_procs[0].id == agent_proc.id

    live_by_env_procs = yield data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 1
    assert live_by_env_procs[0].id == agent_proc.id

    yield agent_proc.update_fields(expired=datetime.datetime.now())

    live_procs = yield data.AgentProcess.get_live()
    assert len(live_procs) == 0

    live_by_env_procs = yield data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 0

    yield agent_proc.delete_cascade()

    assert (yield data.AgentProcess.get_by_id(agent_proc.id)) is None
    assert (yield data.AgentInstance.get_by_id(agi1.id)) is None
    assert (yield data.AgentInstance.get_by_id(agi2.id)) is None


@pytest.mark.gen_test
def test_agent_instance(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    yield agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.id, name=agi1_name, tid=env.id)
    yield agi1.insert()
    agi2_name = "agi2"
    agi2 = data.AgentInstance(process=agent_proc.id, name=agi2_name, tid=env.id)
    yield agi2.insert()

    active_instances = yield data.AgentInstance.active()
    assert len(active_instances) == 2
    assert agi1.id in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = yield data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi1.id
    current_instances = yield data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id

    yield agi1.update_fields(expired=datetime.datetime.now())

    active_instances = yield data.AgentInstance.active()
    assert len(active_instances) == 1
    assert agi1.id not in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = yield data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 0
    current_instances = yield data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id


@pytest.mark.gen_test
def test_agent(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    yield agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.id, name=agi1_name, tid=env.id)
    yield agi1.insert()

    agent1 = data.Agent(environment=env.id, name="agi1_agent1", last_failover=datetime.datetime.now(), paused=False,
                        primary=agi1.id)
    yield agent1.insert()
    agent2 = data.Agent(environment=env.id, name="agi1_agent2", paused=False)
    yield agent2.insert()
    agent3 = data.Agent(environment=env.id, name="agi1_agent3", paused=True)
    yield agent3.insert()

    agents = yield data.Agent.get_list()
    assert len(agents) == 3
    for agent in agents:
        assert agent.id in [agent1.id, agent2.id, agent3.id]

    for agent in [agent1, agent2, agent3]:
        retrieved_agent = yield agent.get(agent.environment, agent.name)
        assert retrieved_agent is not None
        assert retrieved_agent.id == agent.id

    assert agent1.get_status() == "up"
    assert agent2.get_status() == "down"
    assert agent3.get_status() == "paused"

    for agent in [agent1, agent2, agent3]:
        assert agent.to_dict()["state"] == agent.get_status()

    yield agent1.update_fields(paused=True)
    assert agent1.get_status() == "paused"

    yield agent2.update_fields(primary=agi1.id)
    assert agent2.get_status() == "up"

    yield agent3.update_fields(paused=False)
    assert agent3.get_status() == "down"

    primary_instance = yield data.AgentInstance.get_by_id(agent1.primary)
    primary_process = yield data.AgentProcess.get_by_id(primary_instance.process)
    assert primary_process.id == agent_proc.id


@pytest.mark.gen_test
def test_config_model(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(),
                                 total=1, version_info={})
    yield cm.insert()

    # create resources
    key = "std::File[agent1,path=/etc/motd]"
    res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version, attributes={"path": "/etc/motd"})
    yield res1.insert()

    agents = yield data.ConfigurationModel.get_agents(env.id, version)
    assert len(agents) == 1
    assert "agent1" in agents


@pytest.mark.gen_test
def test_model_list(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    for version in range(1, 20):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        yield cm.insert()

    versions = yield data.ConfigurationModel.get_versions(env.id, 0, 1)
    assert len(versions) == 1
    assert versions[0].version == 19

    versions = yield data.ConfigurationModel.get_versions(env.id, 1, 1)
    assert len(versions) == 1
    assert versions[0].version == 18

    versions = yield data.ConfigurationModel.get_versions(env.id)
    assert len(versions) == 19
    assert versions[0].version == 19
    assert versions[-1].version == 1

    versions = yield data.ConfigurationModel.get_versions(env.id, 10)
    assert len(versions) == 9
    assert versions[0].version == 9
    assert versions[-1].version == 1


@pytest.mark.gen_test
def test_model_get_latest_version(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    cms = []
    for version in range(1, 5):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        yield cm.insert()
        cms.append(cm)

    latest_version = yield data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version is None

    yield cms[1].update_fields(released=True)
    latest_version = yield data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 2

    yield cms[3].update_fields(released=True)
    latest_version = yield data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 4


@pytest.mark.gen_test
def test_model_set_ready(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                 version_info={})
    yield cm.insert()

    amount_of_done_resources = cm.done
    assert amount_of_done_resources == 0

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    yield resource.insert()

    assert cm.done == 0

    yield data.ConfigurationModel.set_ready(env.id, version, resource.id, resource.resource_id,
                                            const.ResourceState.deployed)

    cm = yield data.ConfigurationModel.get_by_id(cm.id)
    assert cm.done == 1


@pytest.mark.gen_test
def test_model_delete_cascade(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                 version_info={})
    yield cm.insert()

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    yield resource.insert()

    code = data.Code(version=version, resource="std::File", environment=env.id)
    yield code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    yield unknown_parameter.insert()

    yield cm.delete_cascade()

    assert (yield data.ConfigurationModel.get_by_id(cm.id)) is None
    assert (yield data.Resource.get_by_id(resource.id)) is None
    assert (yield data.Code.get_by_id(code.id)) is None
    assert (yield data.UnknownParameter.get_by_id(unknown_parameter.id)) is None


@pytest.mark.gen_test
def test_undeployable_cache_lazy(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=5, version_info={},
                                  released=False, deployed=False)
    yield cm1.insert()
    yield populate_model(env.id, version)

    assert cm1.undeployable is None

    undep = yield cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]

    assert cm1.undeployable is not None

    undep = yield cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]

    cm1 = yield data.ConfigurationModel.get_version(env.id, version)

    assert cm1.undeployable is not None

    undep = yield cm1.get_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 3]


@pytest.mark.gen_test
def test_undeployable_skip_cache_lazy(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = 2
    cm1 = data.ConfigurationModel(environment=env.id,
                                  version=version,
                                  date=datetime.datetime.now(),
                                  total=5,
                                  version_info={},
                                  released=False, deployed=False)
    yield cm1.insert()
    yield populate_model(env.id, version)

    assert cm1.skipped_for_undeployable is None

    undep = yield cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]

    assert cm1.skipped_for_undeployable is not None

    undep = yield cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]

    cm1 = yield data.ConfigurationModel.get_version(env.id, version)

    assert cm1.skipped_for_undeployable is not None

    undep = yield cm1.get_skipped_for_undeployable()
    assert undep == ["std::File[agent1,path=/tmp/%d]" % 4, "std::File[agent1,path=/tmp/%d]" % 5]


@gen.coroutine
def populate_model(env_id, version):

    def get_path(n):
        return "/tmp/%d" % n

    def get_id(n):
        return "std::File[agent1,path=/tmp/%d],v=%s" % (n, version)

    def get_resource(n, depends, status=const.ResourceState.available):
        requires = [get_id(z) for z in depends]
        return data.Resource.new(environment=env_id, resource_version_id=get_id(n),
                                 status=status,
                                 attributes={"path": get_path(n),
                                             "purge_on_delete": False,
                                             "purged": False,
                                             "requires": requires})

    res1 = get_resource(1, [])
    yield res1.insert()

    res2 = get_resource(2, [1])
    yield res2.insert()

    res3 = get_resource(3, [], const.ResourceState.undefined)
    yield res3.insert()

    res4 = get_resource(4, [3])
    yield res4.insert()

    res5 = get_resource(5, [4])
    yield res5.insert()


@pytest.mark.gen_test
def test_resource_purge_on_delete(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=2, version_info={},
                                  released=True, deployed=True)
    yield cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    yield res11.insert()

    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent2,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    yield res12.insert()

    # model 2 (multiple undeployed versions)
    while version < 10:
        version += 1
        cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                      version_info={}, released=False, deployed=False)
        yield cm2.insert()

        res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent5,path=/etc/motd],v=%s" % version,
                                  status=const.ResourceState.available,
                                  attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        yield res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    yield cm3.insert()

    to_purge = yield data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.gen_test
def test_issue_422(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={},
                                  released=True, deployed=True)
    yield cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    yield res11.insert()

    # model 2 (multiple undeployed versions)
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    yield cm2.insert()

    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    yield res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    yield cm3.insert()

    to_purge = yield data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.gen_test
def test_get_latest_resource(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    key = "std::File[agent1,path=/etc/motd]"
    assert (yield data.Resource.get_latest_version(env.id, key)) is None

    res11 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=1", status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    yield res11.insert()

    res12 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=2", status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    yield res12.insert()

    res = yield data.Resource.get_latest_version(env.id, key)
    assert res.model == 2


@pytest.mark.gen_test
def test_get_resources(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    resource_ids = []
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/tmp/file%d],v=1" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        yield res.insert()
        resource_ids.append(res.resource_version_id)

    resources = yield data.Resource.get_resources(env.id, resource_ids)
    assert len(resources) == len(resource_ids)
    assert sorted([x.resource_version_id for x in resources]) == sorted(resource_ids)

    resources = yield data.Resource.get_resources(env.id, [resource_ids[0], "abcd"])
    assert len(resources) == 1

    resources = yield data.Resource.get_resources(env.id, [])
    assert len(resources) == 0


@pytest.mark.gen_test
def test_get_resources_for_version(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    resource_ids_version_one = []
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/tmp/file%d],v=1" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        yield res.insert()
        resource_ids_version_one.append(res.resource_version_id)

    resource_ids_version_two = []
    for i in range(11, 21):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent2,path=/tmp/file%d],v=2" % i,
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        yield res.insert()
        resource_ids_version_two.append(res.resource_version_id)

    resources = yield data.Resource.get_resources_for_version(env.id, 1)
    assert len(resources) == 10
    assert sorted(resource_ids_version_one) == sorted([x.resource_version_id for x in resources])
    resources = yield data.Resource.get_resources_for_version(env.id, 2)
    assert len(resources) == 10
    assert sorted(resource_ids_version_two) == sorted([x.resource_version_id for x in resources])
    resources = yield data.Resource.get_resources_for_version(env.id, 3)
    assert resources == []


@pytest.mark.gen_test
def test_escaped_resources(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    routes = {"8.0.0.0/8": "1.2.3.4", "0.0.0.0/0": "127.0.0.1"}
    res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,name=router],v=1",
                            status=const.ResourceState.deployed,
                            attributes={"name": "router", "purge_on_delete": True, "purged": False, "routes": routes})
    yield res.insert()
    resource_id = res.resource_version_id

    resources = yield data.Resource.get_resources(env.id, [resource_id])
    assert len(resources) == 1

    assert resources[0].attributes["routes"] == routes


@pytest.mark.gen_test
def test_resource_provides(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res1.provides.append(res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []

    yield res1.insert()
    yield res2.insert()

    res1 = yield data.Resource.get(env.id, res1.resource_version_id)
    res2 = yield data.Resource.get(env.id, res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []


@pytest.mark.gen_test
def test_resources_report(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={},
                                  released=True, deployed=True)
    yield cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file1"})
    yield res11.insert()
    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file2"})
    yield res12.insert()

    # model 2
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    yield cm2.insert()
    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/file1"})
    yield res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    yield cm3.insert()

    res31 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 14, 30),
                              attributes={"path": "/etc/file2"})
    yield res31.insert()

    report = yield data.Resource.get_resources_report(env.id)
    assert len(report) == 2
    report_as_map = {x["resource_id"]: x for x in report}
    assert "std::File[agent1,path=/etc/file1]" in report_as_map
    assert "std::File[agent1,path=/etc/file2]" in report_as_map

    assert report_as_map["std::File[agent1,path=/etc/file1]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["deployed_version"] == 1
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["latest_version"] == 2
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 12, 30)
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["agent"] == "agent1"

    assert report_as_map["std::File[agent1,path=/etc/file1]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["deployed_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["latest_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 14, 30)
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["agent"] == "agent1"


@pytest.mark.gen_test
def test_resource_get_requires(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd"})
    yield res1.insert()
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "requires": [res1.resource_version_id]})
    yield res2.insert()

    assert (yield data.Resource.get_requires(env.id, 1, res2.resource_version_id)) == []
    resources = yield data.Resource.get_requires(env.id, 1, res1.resource_version_id)
    assert len(resources) == 1
    assert resources[0].id == res2.id


@pytest.mark.gen_test
def test_resource_get_with_state(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    yield cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                              attributes={"path": "/etc/file1", "state_id": "test11"})
    yield res11.insert()
    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file1"})
    yield res12.insert()
    res13 = data.Resource.new(environment=env.id,
                              resource_version_id="std::File[agent1,path=/etc/file3],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file1", "state_id": "test13"})
    yield res13.insert()
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    yield cm2.insert()

    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                              attributes={"path": "/etc/file1", "state_id": "test"})
    yield res21.insert()
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    yield cm3.insert()

    res31 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file3],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                              attributes={"path": "/etc/file1"})
    yield res31.insert()

    resources = yield data.Resource.get_with_state(env.id, 1)
    assert len(resources) == 2
    assert sorted([res11.id, res13.id]) == sorted([x.id for x in resources])
    resources = yield data.Resource.get_with_state(env.id, 2)
    assert len(resources) == 1
    assert res21.id == resources[0].id
    resources = yield data.Resource.get_with_state(env.id, 3)
    assert len(resources) == 0


@pytest.mark.gen_test
def test_resources_delete_cascade(data_module):
    project = data.Project(name="test")
    yield project.insert()
    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    yield cm.insert()

    res1 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    yield res1.insert()
    res2 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    yield res2.insert()
    resource_action1 = data.ResourceAction(environment=env.id, resource_version_ids=[res1.resource_version_id],
                                           action_id=uuid.uuid4(), action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    yield resource_action1.insert()
    resource_action2 = data.ResourceAction(environment=env.id, resource_version_ids=[res2.resource_version_id],
                                           action_id=uuid.uuid4(), action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    yield resource_action2.insert()

    yield res1.delete_cascade()

    assert (yield data.Resource.get_by_id(res1.id)) is None
    assert (yield data.ResourceAction.get_by_id(resource_action1.id)) is None
    resource = yield data.Resource.get_by_id(res2.id)
    assert resource is not None
    assert resource.id == res2.id
    resource_action = yield data.ResourceAction.get_by_id(resource_action2.id)
    assert resource_action is not None
    assert resource_action.id == resource_action2.id


@pytest.mark.gen_test
def test_resource_action(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    action_id = uuid.uuid4()
    resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[], action_id=action_id,
                                          action=const.ResourceAction.deploy, started=datetime.datetime.now())
    yield resource_action.insert()

    resource_action.add_changes({"rid": {"field1": {"old": "a", "new": "b"}, "field2": {}}})
    yield resource_action.save()

    resource_action.add_changes({"rid": {"field2": {"old": "c", "new": "d"}, "field3": {}}})
    yield resource_action.save()

    resource_action.add_logs([{}, {}])
    yield resource_action.save()

    resource_action.add_logs([{}, {}])
    yield resource_action.save()

    ra = yield data.ResourceAction.get_by_id(resource_action.id)
    assert len(ra.changes["rid"]) == 3
    assert len(ra.messages) == 4

    assert ra.changes["rid"]["field1"]["old"] == "a"
    assert ra.changes["rid"]["field1"]["new"] == "b"
    assert ra.changes["rid"]["field2"]["old"] == "c"
    assert ra.changes["rid"]["field2"]["new"] == "d"
    assert ra.changes["rid"]["field3"] == {}


@pytest.mark.gen_test
def test_resource_action_get_logs(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    for i in range(1, 11):
        action_id = uuid.uuid4()
        resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[uuid.UUID(int=1)], action_id=action_id,
                                              action=const.ResourceAction.deploy, started=datetime.datetime.now())
        yield resource_action.insert()
        resource_action.add_logs(["test" + str(i)])
        yield resource_action.save()

    action_id = uuid.uuid4()
    resource_action = data.ResourceAction(environment=env.id, resource_version_ids=[uuid.UUID(int=1)], action_id=action_id,
                                          action=const.ResourceAction.dryrun, started=datetime.datetime.now())
    yield resource_action.insert()
    resource_action.add_logs(["log", "message"])
    yield resource_action.save()

    resource_actions = yield data.ResourceAction.get_log(env.id, uuid.UUID(int=1))
    assert len(resource_actions) == 11
    for i in range(len(resource_actions)):
        action = resource_actions[i]
        if i == 0:
            assert action.action == const.ResourceAction.dryrun
        else:
            assert action.action == const.ResourceAction.deploy
    resource_actions = yield data.ResourceAction.get_log(env.id, uuid.UUID(int=1), const.ResourceAction.dryrun.name)
    assert len(resource_actions) == 1
    assert resource_actions[0].action == const.ResourceAction.dryrun
    assert sorted(resource_actions[0].messages) == sorted(["log", "message"])
    resource_actions = yield data.ResourceAction.get_log(env.id, uuid.UUID(int=1), const.ResourceAction.deploy.name, limit=2)
    assert len(resource_actions) == 2
    assert resource_actions[0].messages == ["test10"]
    assert resource_actions[1].messages == ["test9"]


@pytest.mark.gen_test
def test_data_document_recursion(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    now = datetime.datetime.now()
    ra = data.ResourceAction(environment=env.id, resource_version_ids=["id"], action_id=uuid.uuid4(),
                             action=const.ResourceAction.store, started=now, finished=now,
                             messages=[data.LogLine.log(logging.INFO, "Successfully stored version %(version)d",
                                                        version=2)])
    yield ra.insert()


@pytest.mark.gen_test
def test_snapshot(data_module):
    project = data.Project(name="test")
    yield project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    yield env.insert()

    snap = data.Snapshot(environment=env.id, model=1, name="a", started=datetime.datetime.now(), resources_todo=1)
    yield snap.insert()

    s = yield data.Snapshot.get_by_id(snap.id)
    yield s.resource_updated(10)
    assert s.resources_todo == 0
    assert s.total_size == 10
    assert s.finished is not None

    s = yield data.Snapshot.get_by_id(snap.id)
    assert s.resources_todo == 0
    assert s.total_size == 10
    assert s.finished is not None

    yield s.delete_cascade()
    result = yield data.Snapshot.get_list()
    assert len(result) == 0



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import socket
import time
import inspect
import urllib
import uuid
import json
import re
from datetime import datetime
from collections import defaultdict
import enum
import warnings
import io
import gzip

import tornado
from tornado import gen, web
from inmanta import methods, const, execute
from inmanta import config as inmanta_config
from tornado.httpclient import HTTPRequest, AsyncHTTPClient, HTTPError
from tornado.ioloop import IOLoop
import jwt

from inmanta.util import Scheduler


LOGGER = logging.getLogger(__name__)
INMANTA_MT_HEADER = "X-Inmanta-tid"

"""

RestServer  => manages tornado/handlers, marshalling, dispatching, and endpoints

ServerSlice => contributes handlers and methods

ServerSlice.server [1] -- RestServer.endpoints [1:]

"""


# Util functions
def custom_json_encoder(o):
    """
        A custom json encoder that knows how to encode other types commonly used by Inmanta
    """
    if isinstance(o, uuid.UUID):
        return str(o)

    if isinstance(o, datetime):
        return o.isoformat()

    if hasattr(o, "to_dict"):
        return o.to_dict()

    if isinstance(o, enum.Enum):
        return o.name

    if isinstance(o, Exception):
        # Logs can push exceptions through RPC. Return a string representation.
        return str(o)

    if isinstance(o, execute.util.Unknown):
        return const.UNKNOWN_STRING

    LOGGER.error("Unable to serialize %s", o)
    raise TypeError(repr(o) + " is not JSON serializable")


def json_encode(value):
    # see json_encode in tornado.escape
    return json.dumps(value, default=custom_json_encoder).replace("</", "<\\/")


def gzipped_json(value):
    value = json_encode(value)
    if len(value) < web.GZipContentEncoding.MIN_LENGTH:
        return False, value

    gzip_value = io.BytesIO()
    gzip_file = gzip.GzipFile(mode="w", fileobj=gzip_value, compresslevel=web.GZipContentEncoding.GZIP_LEVEL)

    gzip_file.write(value.encode())
    gzip_file.close()

    return True, gzip_value.getvalue()


def sh(msg, max_len=10):
    if len(msg) < max_len:
        return msg
    return msg[0:max_len - 3] + "..."


def encode_token(client_types, environment=None, idempotent=False, expire=None):
    cfg = inmanta_config.AuthJWTConfig.get_sign_config()

    payload = {
        "iss": cfg.issuer,
        "aud": [cfg.audience],
        const.INMANTA_URN + "ct": ",".join(client_types),
    }

    if not idempotent:
        payload["iat"] = int(time.time())

        if cfg.expire > 0:
            payload["exp"] = int(time.time() + cfg.expire)
        elif expire is not None:
            payload["exp"] = int(time.time() + expire)

    if environment is not None:
        payload[const.INMANTA_URN + "env"] = environment

    return jwt.encode(payload, cfg.key, cfg.algo).decode()


def decode_token(token):
    try:
        # First decode the token without verification
        header = jwt.get_unverified_header(token)
        payload = jwt.decode(token, verify=False)
    except Exception as e:
        raise UnauhorizedError("Unable to decode provided JWT bearer token.")

    if "iss" not in payload:
        raise UnauhorizedError("Issuer is required in token to validate.")

    cfg = inmanta_config.AuthJWTConfig.get_issuer(payload["iss"])
    if cfg is None:
        raise UnauhorizedError("Unknown issuer for token")

    alg = header["alg"].lower()
    if alg == "hs256":
        key = cfg.key
    elif alg == "rs256":
        if "kid" not in header:
            raise UnauhorizedError("A kid is required for RS256")
        kid = header["kid"]
        if kid not in cfg.keys:
            raise UnauhorizedError("The kid provided in the token does not match a known key. Check the jwks_uri or try "
                                   "restarting the server to load any new keys.")

        key = cfg.keys[kid]
    else:
        raise UnauhorizedError("Algorithm %s is not supported." % alg)

    try:
        payload = jwt.decode(token, key, audience=cfg.audience, algorithms=[cfg.algo])
        ct_key = const.INMANTA_URN + "ct"
        payload[ct_key] = [x.strip() for x in payload[ct_key].split(",")]
    except Exception as e:
        raise UnauhorizedError(*e.args)

    return payload


def authorize_request(auth_data, metadata, message, config):
    """
        Authorize a request based on the given data
    """
    if auth_data is None:
        return

    # Enforce environment restrictions
    env_key = const.INMANTA_URN + "env"
    if env_key in auth_data:
        if env_key not in metadata:
            raise UnauhorizedError("The authorization token is scoped to a specific environment.")

        if metadata[env_key] != "all" and auth_data[env_key] != metadata[env_key]:
            raise UnauhorizedError("The authorization token is not valid for the requested environment.")

    # Enforce client_types restrictions
    ok = False
    ct_key = const.INMANTA_URN + "ct"
    for ct in auth_data[ct_key]:
        if ct in config[0]["client_types"]:
            ok = True

    if not ok:
        raise UnauhorizedError("The authorization token does not have a valid client type for this call." +
                               " (%s provided, %s expected" % (auth_data[ct_key], config[0]["client_types"]))

    return


# API
class UnauhorizedError(Exception):
    pass


class Result(object):
    """
        A result of a method call
    """

    def __init__(self, multiple=False, code=0, result=None):
        self._multiple = multiple
        if multiple:
            self._result = []
            if result is not None:
                self._result.append(result)
        else:
            self._result = result
        self.code = code
        self._callback = None

    def add_result(self, result):
        """
        Add a new result to an instance

        :param result: The result to store
        """
        assert(self._multiple)
        self._result.append(result)
        if self._callback:
            self._callback(self)

    def get_result(self):
        """
            Only when the result is marked as available the result can be returned
        """
        if self.available():
            return self._result
        raise Exception("The result is not yet available")

    def set_result(self, value):
        if not self.available():
            assert(not self._multiple)
            self._result = value
            if self._callback:
                self._callback(self)

    def available(self):
        if self._multiple:
            return len(self._result) > 0 is not None or self.code > 0
        else:
            return self._result is not None or self.code > 0

    def wait(self, timeout=60):
        """
            Wait for the result to become available
        """
        count = 0
        while count < timeout:
            time.sleep(0.1)
            count += 0.1

    result = property(get_result, set_result)

    def callback(self, fnc):
        """
            Set a callback function that is to be called when the result is ready. When multiple
            results are expected, the callback is called for each result.
        """
        self._callback = fnc


# Tornado Interface


# Shared
class RESTBase(object):

    def _create_base_url(self, properties, msg=None, versioned=True):
        """
            Create a url for the given protocol properties
        """
        url = "/api/v1" if versioned else ""
        if "id" in properties and properties["id"]:
            if msg is None:
                url += "/%s/(?P<id>[^/]+)" % properties["method_name"]
            else:
                url += "/%s/%s" % (properties["method_name"], urllib.parse.quote(str(msg["id"]), safe=""))

        elif "index" in properties and properties["index"]:
            url += "/%s" % properties["method_name"]
        else:
            url += "/%s" % properties["method_name"]

        return url

    def _decode(self, body):
        """
            Decode a response body
        """
        if body is not None and len(body) > 0:
            body = json.loads(tornado.escape.to_basestring(body))
        else:
            body = None

        return body

    @gen.coroutine
    def _execute_call(self, kwargs, http_method, config, message, request_headers, auth=None):
        if "api_version" in config[0] and config[0]["api_version"] is None:
            warnings.warn("Using an unversioned API method will be removed in the next release", DeprecationWarning)
            LOGGER.warning("Using an unversioned API method will be removed in the next release")

        headers = {"Content-Type": "application/json"}
        try:
            if kwargs is None or config is None:
                raise Exception("This method is unknown! This should not occur!")
            # create message that contains all arguments (id, query args and body)
            if "id" in kwargs and (message is None or "id" not in message):
                message["id"] = kwargs["id"]

            # validate message against config
            if "id" in config[0] and config[0]["id"] and "id" not in message:
                return self.return_error_msg(500, "Invalid request. It should contain an id in the url.", headers)

            validate_sid = config[0]["validate_sid"]
            if validate_sid:
                if 'sid' not in message:
                    return self.return_error_msg(500,
                                                 "This is an agent to server call, it should contain an agent session id",
                                                 headers)
                elif not self.validate_sid(message['sid']):
                    return self.return_error_msg(500, "The sid %s is not valid." % message['sid'], headers)

            # validate message against the arguments
            argspec = inspect.getfullargspec(config[2])
            args = argspec.args

            if "self" in args:
                args.remove("self")

            all_fields = set(message.keys())
            if argspec.defaults is not None:
                defaults_start = len(args) - len(argspec.defaults)
            else:
                defaults_start = -1

            metadata = {}
            for i in range(len(args)):
                arg = args[i]

                opts = {}
                # handle defaults and header mapping
                if arg in config[0]["arg_options"]:
                    opts = config[0]["arg_options"][arg]

                    if "header" in opts:
                        message[arg] = request_headers[opts["header"]]
                        if "reply_header" in opts and opts["reply_header"]:
                            headers[opts["header"]] = message[arg]
                    all_fields.add(arg)

                if arg not in message:
                    if (defaults_start >= 0 and (i - defaults_start) < len(argspec.defaults)):
                        message[arg] = argspec.defaults[i - defaults_start]
                    else:
                        return self.return_error_msg(500, "Invalid request. Field '%s' is required." % arg, headers)

                else:
                    all_fields.remove(arg)

                # validate the type
                if arg in argspec.annotations:
                    arg_type = argspec.annotations[arg]
                    if message[arg] is not None and not isinstance(message[arg], arg_type):
                        try:
                            if arg_type == datetime:
                                message[arg] = datetime.strptime(message[arg], "%Y-%m-%dT%H:%M:%S.%f")

                            elif issubclass(arg_type, enum.Enum):
                                message[arg] = arg_type[message[arg]]

                            elif arg_type == bool:
                                message[arg] = inmanta_config.is_bool(message[arg])

                            else:
                                message[arg] = arg_type(message[arg])

                        except (ValueError, TypeError):
                            error_msg = ("Invalid type for argument %s. Expected %s but received %s" %
                                         (arg, arg_type, message[arg].__class__))
                            LOGGER.exception(error_msg)
                            return self.return_error_msg(500, error_msg, headers)

                # execute any getters that are defined
                if "getter" in opts:
                    try:
                        result = yield opts["getter"](message[arg], metadata)
                        message[arg] = result
                    except methods.HTTPException as e:
                        LOGGER.exception("Failed to use getter for arg %s", arg)
                        return self.return_error_msg(e.code, e.message, headers)

            if config[0]["agent_server"]:
                if 'sid' in all_fields:
                    del message['sid']
                    all_fields.remove('sid')

            if len(all_fields) > 0 and argspec.varkw is None:
                return self.return_error_msg(500, ("Request contains fields %s " % all_fields) +
                                             "that are not declared in method and no kwargs argument is provided.", headers)

            LOGGER.debug("Calling method %s(%s)", config[1][1], ", ".join(["%s='%s'" % (name, sh(str(value)))
                                                                           for name, value in message.items()]))
            method_call = getattr(config[1][0], config[1][1])

            if hasattr(method_call, "__protocol_mapping__"):
                for k, v in method_call.__protocol_mapping__.items():
                    if v in message:
                        message[k] = message[v]
                        del message[v]

            try:
                authorize_request(auth, metadata, message, config)
            except UnauhorizedError as e:
                return self.return_error_msg(403, e.args[0], headers)

            result = yield method_call(**message)

            if result is None:
                raise Exception("Handlers for method calls should at least return a status code. %s on %s" % config[1])

            reply = None
            if isinstance(result, tuple):
                if len(result) == 2:
                    code, reply = result
                else:
                    raise Exception("Handlers for method call can only return a status code and a reply")

            else:
                code = result

            if reply is not None:
                if "reply" in config[0] and config[0]:
                    LOGGER.debug("%s returned %d: %s", config[1][1], code, sh(str(reply), 70))
                    return reply, headers, code

                else:
                    LOGGER.warn("Method %s returned a result although it is has not reply!")

            return None, headers, code

        except Exception as e:
            LOGGER.exception("An exception occured during the request.")
            return self.return_error_msg(500, "An exception occured: " + str(e.args), headers)


# Client side
class RESTTransport(RESTBase):
    """"
        A REST (json body over http) transport. Only methods that operate on resource can use all
        HTTP verbs. For other methods the POST verb is used.
    """
    __transport_name__ = "rest"

    def __init__(self, endpoint, connection_timout=120):
        self.__end_point = endpoint
        self.daemon = True
        self._connected = False
        self.set_connected()
        self._handlers = []
        self.token = inmanta_config.Config.get(self.id, "token", None)
        self.connection_timout = connection_timout
        self.headers = set()

    endpoint = property(lambda x: x.__end_point)

    def get_id(self):
        """
            Returns a unique id for a transport on an endpoint
        """
        return "%s_%s_transport" % (self.__end_point.name, self.__class__.__transport_name__)

    id = property(get_id)

    def start_client(self):
        """
            Start this transport as client
        """
        self.start()

    def stop_client(self):
        """
            Stop this transport as client
        """
        self.stop()

    def start(self):
        """
            Start the transport as a new thread
        """
        pass

    def stop(self):
        """
            Stop the transport
        """
        self._connected = False

    def set_connected(self):
        """
            Mark this transport as connected
        """
        LOGGER.debug("Transport %s is connected", self.get_id())
        self._connected = True

    def is_connected(self):
        """
            Is this transport connected
        """
        return self._connected

    def create_op_mapping(self):
        """
            Build a mapping between urls, ops and methods
        """
        url_map = defaultdict(dict)
        headers = set()
        for method, method_handlers in self.endpoint.__methods__.items():
            properties = method.__protocol_properties__
            call = (self.endpoint, method_handlers[0])

            if "arg_options" in properties:
                for opts in properties["arg_options"].values():
                    if "header" in opts:
                        headers.add(opts["header"])

            url = self._create_base_url(properties)
            properties["api_version"] = "1"
            url_map[url][properties["operation"]] = (properties, call, method.__wrapped__)

            url = self._create_base_url(properties, versioned=False)
            properties = properties.copy()
            properties["api_version"] = None
            url_map[url][properties["operation"]] = (properties, call, method.__wrapped__)

        headers.add("Authorization")
        self.headers = headers
        return url_map

    def match_call(self, url, method):
        """
            Get the method call for the given url and http method
        """
        url_map = self.create_op_mapping()
        for url_re, handlers in url_map.items():
            if not url_re.endswith("$"):
                url_re += "$"
            match = re.match(url_re, url)
            if match and method in handlers:
                return match.groupdict(), handlers[method]

        return None, None

    def _get_client_config(self):
        """
            Load the configuration for the client
        """
        LOGGER.debug("Getting config in section %s", self.id)

        port = inmanta_config.Config.get(self.id, "port", 8888)

        host = inmanta_config.Config.get(self.id, "host", "localhost")

        if inmanta_config.Config.getboolean(self.id, "ssl", False):
            protocol = "https"
        else:
            protocol = "http"

        return "%s://%s:%d" % (protocol, host, port)

    def build_call(self, properties, args, kwargs={}):
        """
            Build a call from the given arguments. This method returns the url, headers, method and body for the call.
        """
        method = properties["operation"]

        # create the message
        msg = kwargs

        # map the argument in arg to names
        argspec = inspect.getfullargspec(properties["method"])
        for i in range(len(args)):
            msg[argspec.args[i + 1]] = args[i]

        url = self._create_base_url(properties, msg)

        headers = {}

        for arg_name in list(msg.keys()):
            if isinstance(msg[arg_name], enum.Enum):  # Handle enum values "special"
                msg[arg_name] = msg[arg_name].name

            if arg_name in properties["arg_options"]:
                opts = properties["arg_options"][arg_name]
                if "header" in opts:
                    headers[opts["header"]] = str(msg[arg_name])
                    del msg[arg_name]

        if not (method == "POST" or method == "PUT" or method == "PATCH"):
            qs_map = msg.copy()
            if "id" in qs_map:
                del qs_map["id"]

            # encode arguments in url
            if len(qs_map) > 0:
                url += "?" + urllib.parse.urlencode(qs_map)

            body = None
        else:
            body = msg

        return url, method, headers, body

    @gen.coroutine
    def call(self, properties, args, kwargs={}):
        url, method, headers, body = self.build_call(properties, args, kwargs)

        url_host = self._get_client_config()
        url = url_host + url

        if self.token is not None:
            headers["Authorization"] = "Bearer " + self.token

        if body is not None:
            zipped, body = gzipped_json(body)
            if zipped:
                headers["Content-Encoding"] = "gzip"

        ca_certs = inmanta_config.Config.get(self.id, "ssl_ca_cert_file", None)
        LOGGER.debug("Calling server %s %s", method, url)

        try:
            request = HTTPRequest(url=url, method=method, headers=headers, body=body, connect_timeout=self.connection_timout,
                                  request_timeout=120, ca_certs=ca_certs, decompress_response=True)
            client = AsyncHTTPClient()
            response = yield client.fetch(request)
        except HTTPError as e:
            if e.response is not None and len(e.response.body) > 0:
                try:
                    result = self._decode(e.response.body)
                except ValueError:
                    result = {}
                return Result(code=e.code, result=result)

            return Result(code=e.code, result={"message": str(e)})
        except Exception as e:
            raise e
            return Result(code=500, result={"message": str(e)})

        return Result(code=response.code, result=self._decode(response.body))

    def validate_sid(self, sid):
        return self.endpoint.validate_sid(sid)


class handle(object):  # noqa: H801
    """
        Decorator for subclasses of an endpoint to handle protocol methods

        :param method A subclass of method that defines the method
    """

    def __init__(self, method, **kwargs):
        self.method = method
        self.mapping = kwargs

    def __call__(self, function):
        """
            The wrapping
        """
        class_name = self.method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
        module = inspect.getmodule(self.method)
        method_class = getattr(module, class_name)

        if not hasattr(method_class, "__method_name__"):
            raise Exception("%s should have a __method_name__ variable." % method_class)

        if "method_name" not in self.method.__protocol_properties__:
            self.method.__protocol_properties__["method_name"] = method_class.__method_name__

        function.__protocol_method__ = self.method
        function.__protocol_mapping__ = self.mapping
        return function


class EndpointMeta(type):
    """
        Meta class to create endpoints
    """
    def __new__(cls, class_name, bases, dct):
        if "__methods__" not in dct:
            dct["__methods__"] = {}

        total_dict = dct.copy()
        for base in bases:
            total_dict.update(base.__dict__)

        methods = {}
        for name, attr in total_dict.items():
            if name[0:2] != "__" and hasattr(attr, "__protocol_method__"):
                if attr.__protocol_method__ in methods:
                    raise Exception("Unable to register multiple handlers for the same method.")

                methods[attr.__protocol_method__] = (name, attr)

        dct["__methods__"] = methods

        return type.__new__(cls, class_name, bases, dct)


class Endpoint(object):
    """
        An end-point in the rpc framework
    """

    def __init__(self, io_loop, name):
        self._name = name
        self._node_name = inmanta_config.nodename.get()
        self._end_point_names = []
        self._io_loop = io_loop

    def add_future(self, future):
        """
            Add a future to the ioloop to be handled, but do not require the result.
        """
        def handle_result(f):
            try:
                f.result()
            except Exception as e:
                LOGGER.exception("An exception occurred while handling a future: %s", str(e))

        self._io_loop.add_future(future, handle_result)

    def get_end_point_names(self):
        return self._end_point_names

    def add_end_point_name(self, name):
        """
            Add an additional name to this endpoint to which it reacts and sends out in heartbeats
        """
        LOGGER.debug("Adding '%s' as endpoint", name)
        self._end_point_names.append(name)

    def clear_end_points(self):
        """
            Clear all endpoints
        """
        self._end_point_names = []

    name = property(lambda self: self._name)
    end_point_names = property(get_end_point_names)

    def _get_hostname(self):
        """
            Determine the hostname of this machine
        """
        return socket.gethostname()

    def get_node_name(self):
        return self._node_name

    node_name = property(get_node_name)


class AgentEndPoint(Endpoint, metaclass=EndpointMeta):
    """
        An endpoint for clients that make calls to a server and that receive calls back from the server using long-poll
    """

    def __init__(self, name, io_loop, timeout=120, transport=RESTTransport, reconnect_delay=5):
        super().__init__(io_loop, name)
        self._transport = transport
        self._client = None
        self._sched = Scheduler(self._io_loop)

        self._env_id = None

        self.sessionid = uuid.uuid1()
        self.running = True
        self.server_timeout = timeout
        self.reconnect_delay = reconnect_delay

    def get_environment(self):
        return self._env_id

    environment = property(get_environment)

    def set_environment(self, environment_id: uuid.UUID):
        """
            Set the environment of this agent
        """
        if isinstance(environment_id, str):
            self._env_id = uuid.UUID(environment_id)
        else:
            self._env_id = environment_id

    def start(self):
        """
            Connect to the server and use a heartbeat and long-poll for two-way communication
        """
        assert self._env_id is not None
        LOGGER.log(3, "Starting agent for %s", str(self.sessionid))
        self._client = AgentClient(self.name, self.sessionid, transport=self._transport, timeout=self.server_timeout)
        self._io_loop.add_callback(self.perform_heartbeat)

    def stop(self):
        self.running = False

    @gen.coroutine
    def on_reconnect(self):
        pass

    @gen.coroutine
    def perform_heartbeat(self):
        """
            Start a continuous heartbeat call
        """
        connected = False
        while self.running:
            LOGGER.log(3, "sending heartbeat for %s", str(self.sessionid))
            result = yield self._client.heartbeat(sid=str(self.sessionid),
                                                  tid=str(self._env_id),
                                                  endpoint_names=self.end_point_names,
                                                  nodename=self.node_name)
            LOGGER.log(3, "returned heartbeat for %s", str(self.sessionid))
            if result.code == 200:
                if not connected:
                    connected = True
                    self.add_future(self.on_reconnect())
                if result.result is not None:
                    if "method_calls" in result.result:
                        method_calls = result.result["method_calls"]
                        transport = self._transport(self)

                        for method_call in method_calls:
                            self.dispatch_method(transport, method_call)
            else:
                LOGGER.warning("Heartbeat failed with status %d and message: %s, going to sleep for %d s",
                               result.code, result.result, self.reconnect_delay)
                connected = False
                yield gen.sleep(self.reconnect_delay)

    def dispatch_method(self, transport, method_call):
        LOGGER.debug("Received call through heartbeat: %s %s %s", method_call[
                     "reply_id"], method_call["method"], method_call["url"])
        kwargs, config = transport.match_call(method_call["url"], method_call["method"])

        if config is None:
            msg = "An error occurred during heartbeat method call (%s %s %s): %s" % (
                method_call["reply_id"], method_call["method"], method_call["url"], "No such method")
            LOGGER.error(msg)
            self.add_future(self._client.heartbeat_reply(self.sessionid, method_call["reply_id"],
                                                         {"result": msg, "code": 500}))

        body = {}
        if "body" in method_call and method_call["body"] is not None:
            body = method_call["body"]

        query_string = urllib.parse.urlparse(method_call["url"]).query
        for key, value in urllib.parse.parse_qs(query_string, keep_blank_values=True):
            if len(value) == 1:
                body[key] = value[0].decode("latin-1")
            else:
                body[key] = [v.decode("latin-1") for v in value]

        call_result = self._transport(self)._execute_call(kwargs, method_call["method"], config, body,
                                                          method_call["headers"])

        def submit_result(future):
            if future is None:
                return

            result_body, _, status = future.result()
            if status == 500:
                msg = ""
                if result_body is not None and "message" in result_body:
                    msg = result_body["message"]
                LOGGER.error("An error occurred during heartbeat method call (%s %s %s): %s",
                             method_call["reply_id"], method_call["method"], method_call["url"], msg)
            self._client.heartbeat_reply(self.sessionid, method_call["reply_id"],
                                         {"result": result_body, "code": status})

        self._io_loop.add_future(call_result, submit_result)


class ClientMeta(type):
    """
        A meta class that programs all protocol method call into the Client class.
    """
    def __new__(cls, class_name, bases, dct):
        classes = methods.Method.__subclasses__()  # @UndefinedVariable

        for mcls in classes:
            for attr in dir(mcls):
                attr_fn = getattr(mcls, attr)
                if attr[0:2] != "__" and hasattr(attr_fn, "__wrapped__"):
                    dct[attr] = attr_fn

        return type.__new__(cls, class_name, bases, dct)


def get_method_name(properties):
    method = properties["method"]
    class_name = method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0]
    module = inspect.getmodule(method)
    method_class = getattr(module, class_name)

    if not hasattr(method_class, "__method_name__"):
        raise Exception("%s should have a __method_name__ variable." % method_class)

    return method_class.__method_name__


class Client(Endpoint, metaclass=ClientMeta):
    """
        A client that communicates with end-point based on its configuration
    """

    def __init__(self, name, ioloop=None, transport=RESTTransport):
        if ioloop is None:
            ioloop = IOLoop.current()
        Endpoint.__init__(self, ioloop, name)
        self._transport = transport
        self._transport_instance = None

        LOGGER.debug("Start transport for client %s", self.name)
        tr = self._transport(self)
        self._transport_instance = tr

    @gen.coroutine
    def _call(self, args, kwargs, protocol_properties):
        """
            Execute the rpc call
        """
        protocol_properties["method_name"] = get_method_name(protocol_properties)
        result = yield self._transport_instance.call(protocol_properties, args, kwargs)
        return result


class SyncClient(object):
    """
        A synchronous client that communicates with end-point based on its configuration
    """

    def __init__(self, name, timeout=120):
        self.name = name
        self.timeout = timeout
        self._client = Client(self.name)

    def __getattr__(self, name):
        def async_call(*args, **kwargs):
            method = getattr(self._client, name)

            def method_call():
                return method(*args, **kwargs)

            try:
                return IOLoop.current().run_sync(method_call, self.timeout)
            except TimeoutError:
                raise ConnectionRefusedError()

        return async_call


class AgentClient(Endpoint, metaclass=ClientMeta):
    """
        A client that communicates with end-point based on its configuration
    """

    def __init__(self, name, sid, ioloop=None, transport=RESTTransport, timeout=120):
        if ioloop is None:
            ioloop = IOLoop.current()
        Endpoint.__init__(self, ioloop, name)
        self._transport = transport
        self._transport_instance = None
        self._sid = sid

        LOGGER.debug("Start transport for client %s", self.name)
        tr = self._transport(self, connection_timout=timeout)
        self._transport_instance = tr

    @gen.coroutine
    def _call(self, args, kwargs, protocol_properties):
        """
            Execute the rpc call
        """
        protocol_properties["method_name"] = get_method_name(protocol_properties)

        if 'sid' not in kwargs:
            kwargs['sid'] = self._sid

        result = yield self._transport_instance.call(protocol_properties, args, kwargs)
        return result


class ReturnClient(Client, metaclass=ClientMeta):
    """
        A client that uses a return channel to connect to its destination. This client is used by the server to communicate
        back to clients over the heartbeat channel.
    """

    def __init__(self, name, session):
        super().__init__(name)
        self.session = session

    @gen.coroutine
    def _call(self, args, kwargs, protocol_properties):
        protocol_properties["method_name"] = get_method_name(protocol_properties)
        url, method, headers, body = self._transport_instance.build_call(protocol_properties, args, kwargs)

        call_spec = {"url": url, "method": method, "headers": headers, "body": body}
        timeout = protocol_properties["timeout"]
        try:
            return_value = yield self.session.put_call(call_spec, timeout=timeout)
        except gen.TimeoutError:
            return Result(code=500, result="Call timed out")

        return Result(code=return_value["code"], result=return_value["result"])



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from unittest.mock import Mock
from uuid import uuid4, UUID

import pytest
from inmanta.server.agentmanager import AgentManager
from inmanta import data
from tornado import gen
from inmanta.protocol import Result
from utils import assert_equal_ish, UNKWN


class Collector(object):

    def __init__(self):
        self.values = []

    def __call__(self, arg):
        self.values.append(arg)

    @gen.coroutine
    def proccess(self):
        while len(self.values) > 0:
            x = self.values
            self.values = []
            yield x


@gen.coroutine
def empty_future(*args, **kwargs):
    pass


class MockSession(object):
    """
        An environment that segments agents connected to the server
    """

    def __init__(self, sid, tid, endpoint_names, nodename):
        self._sid = sid
        self.tid = tid
        self.endpoint_names = endpoint_names
        self.nodename = nodename
        self.client = Mock()
        self.client.set_state.side_effect = empty_future

    def get_id(self):
        return self._sid

    id = property(get_id)

    def get_client(self):
        return self.client


@pytest.mark.gen_test(timeout=30)
def test_primary_selection(motor):
    data.use_motor(motor)

    env = data.Environment(name="testenv", project=uuid4())
    yield env.insert()

    yield data.Agent(environment=env.id, name="agent1", paused=True).insert()
    yield data.Agent(environment=env.id, name="agent2", paused=False).insert()
    yield data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures

    @gen.coroutine
    def assert_agent(name: str, state: str, sid: UUID):
        agent = yield data.Agent.get(env.id, name)

        assert agent.get_status() == state
        if state == "paused":
            assert agent.primary is None
            assert agent.paused
        elif state == "down":
            assert agent.primary is None
            assert not agent.paused
        elif state == "up":
            assert agent.primary is not None
            agent_instance = yield data.AgentInstance.get_by_id(agent.primary)
            agent_proc = yield data.AgentProcess.get_by_id(agent_instance.process)
            assert agent_proc.sid == sid
            assert agent.get_status() == "up"

    @gen.coroutine
    def assert_agents(s1, s2, s3, sid1=None, sid2=None, sid3=None):
        yield assert_agent("agent1", s1, sid1)
        yield assert_agent("agent2", s2, sid2)
        yield assert_agent("agent3", s3, sid3)

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    yield futures.proccess()
    assert len(am.sessions) == 1
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    yield futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # expire first
    am.expire(ts1, 100)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # expire second
    am.expire(ts2, 100)
    yield futures.proccess()
    assert len(am.sessions) == 0
    yield assert_agents("paused", "down", "down")


@pytest.mark.gen_test(timeout=30)
def test_api(motor):
    data.use_motor(motor)

    env = data.Environment(name="testenv", project=uuid4())
    yield env.insert()
    env2 = data.Environment(name="testenv2", project=uuid4())
    yield env2.insert()
    yield data.Agent(environment=env.id, name="agent1", paused=True).insert()
    yield data.Agent(environment=env.id, name="agent2", paused=False).insert()
    yield data.Agent(environment=env.id, name="agent3", paused=False).insert()
    yield data.Agent(environment=env2.id, name="agent4", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    # third session
    env3 = uuid4()
    ts3 = MockSession(uuid4(), env3, ["agentx"], "ts3")
    am.new_session(ts3)

    yield futures.proccess()
    assert len(am.sessions) == 3

    code, all_agents = yield am.list_agent_processes(None, None)
    assert code == 200

    shouldbe = {'processes': [{'id': UNKWN, 'first_seen': UNKWN, 'expired': None, 'hostname': 'ts1',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent1', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent2', 'process': UNKWN}],
                               'environment': env.id},
                              {'id': UNKWN, 'first_seen': UNKWN, 'expired': None, 'hostname': 'ts2',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent2', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent3', 'process': UNKWN}],
                               'environment': env.id},
                              {'id': UNKWN, 'first_seen': UNKWN, 'expired': None, 'hostname': 'ts3',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agentx', 'process': UNKWN}],
                               'environment': env3}]}

    assert_equal_ish(shouldbe, all_agents, ['hostname', 'name'])
    agentid = all_agents['processes'][0]['id']

    code, all_agents = yield am.list_agent_processes(env.id, None)
    assert code == 200

    shouldbe = {'processes': [{'id': UNKWN, 'first_seen': UNKWN, 'expired': None, 'hostname': 'ts1',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent1', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent2', 'process': UNKWN}],
                               'environment': env.id},
                              {'id': UNKWN, 'first_seen': UNKWN, 'expired': None, 'hostname': 'ts2',
                               'last_seen': UNKWN, 'endpoints':
                               [{'id': UNKWN, 'name': 'agent3', 'process': UNKWN},
                                {'id': UNKWN, 'name': 'agent2', 'process': UNKWN}],
                               'environment': env.id}]}

    assert_equal_ish(shouldbe, all_agents)

    code, all_agents = yield am.list_agent_processes(env2.id, None)
    assert code == 200

    shouldbe = {'processes': []}

    assert_equal_ish(shouldbe, all_agents)

    @gen.coroutine
    def dummy_status():
        return Result(False, 200, "X")

    ts1.get_client().get_status.side_effect = dummy_status
    report = yield am.get_agent_process_report(agentid)
    assert (200, "X") == report

    report = yield am.get_agent_process_report(uuid4())
    assert 404 == report[0]

    code, all_agents = yield am.list_agents(None)
    assert code == 200
    shouldbe = {'agents': [{'name': 'agent1', 'paused': True, 'last_failover': '', 'primary': '',
                            'environment': env.id, "state": "paused"},
                           {'name': 'agent2', 'paused': False, 'last_failover': UNKWN,
                               'primary': UNKWN, 'environment': env.id, "state": "up"},
                           {'name': 'agent3', 'paused': False, 'last_failover': UNKWN,
                               'primary': UNKWN, 'environment': env.id, "state": "up"},
                           {'name': 'agent4', 'paused': False, 'last_failover': '', 'primary': '',
                            'environment': env2.id, "state": "down"}]}
    assert_equal_ish(shouldbe, all_agents, ['name'])

    code, all_agents = yield am.list_agents(env2)
    assert code == 200
    shouldbe = {
        'agents': [{'name': 'agent4', 'paused': False, 'last_failover': '', 'primary': '',
                    'environment': env2.id, "state": "down"}]}
    assert_equal_ish(shouldbe, all_agents)


@pytest.mark.gen_test(timeout=30)
def test_db_clean(motor):
    data.use_motor(motor)

    env = data.Environment(name="testenv", project=uuid4())
    yield env.insert()
    yield data.Agent(environment=env.id, name="agent1", paused=True).insert()
    yield data.Agent(environment=env.id, name="agent2", paused=False).insert()
    yield data.Agent(environment=env.id, name="agent3", paused=False).insert()

    server = Mock()
    futures = Collector()
    server.add_future.side_effect = futures
    am = AgentManager(server, False)
    am.add_future = futures

    @gen.coroutine
    def assert_agent(name: str, state: str, sid: UUID):
        agent = yield data.Agent.get(env.id, name)
        assert agent.get_status() == state
        if state == "paused":
            assert agent.primary is None
            assert agent.paused
        elif state == "down":
            assert agent.primary is None
            assert not agent.paused
        elif state == "up":
            assert agent.primary is not None
            agent_instance = yield data.AgentInstance.get_by_id(agent.primary)
            agent_proc = yield data.AgentProcess.get_by_id(agent_instance.process)
            assert agent_proc.sid == sid
            assert agent.get_status() == "up"

    @gen.coroutine
    def assert_agents(s1, s2, s3, sid1=None, sid2=None, sid3=None):
        yield assert_agent("agent1", s1, sid1)
        yield assert_agent("agent2", s2, sid2)
        yield assert_agent("agent3", s3, sid3)

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    yield futures.proccess()
    assert len(am.sessions) == 1
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    yield futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # expire first
    am.expire(ts1, 100)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # failover
    am = AgentManager(server, False)
    am.add_future = futures
    yield am.clean_db()

    # one session
    ts1 = MockSession(uuid4(), env.id, ["agent1", "agent2"], "ts1")
    am.new_session(ts1)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts1.get_client().set_state.assert_called_with("agent2", True)
    ts1.get_client().reset_mock()
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # alive
    am.seen(ts1, ["agent1", "agent2"])
    yield futures.proccess()
    assert len(am.sessions) == 1
    yield assert_agents("paused", "up", "down", sid2=ts1.id)

    # second session
    ts2 = MockSession(uuid4(), env.id, ["agent3", "agent2"], "ts2")
    am.new_session(ts2)
    yield futures.proccess()
    assert len(am.sessions) == 2
    ts2.get_client().set_state.assert_called_with("agent3", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts1.id, sid3=ts2.id)

    # expire first
    am.expire(ts1, 100)
    yield futures.proccess()
    assert len(am.sessions) == 1
    ts2.get_client().set_state.assert_called_with("agent2", True)
    ts2.get_client().reset_mock()
    yield assert_agents("paused", "up", "up", sid2=ts2.id, sid3=ts2.id)

    # expire second
    am.expire(ts2, 100)
    yield futures.proccess()
    assert len(am.sessions) == 0
    yield assert_agents("paused", "down", "down")
